import java.util.Collections;
import java.util.Comparator;
import java.util.Random;
import java.util.Vector;


/** Classe responsvel por armazenar todos os indivduos da populao, alm de escolher
 * atravs da roleta quais sero reproduzidos
 * <BR><BR>
 * Trabalho de Algoritmos Genticos <BR> 
 * Disciplina de Inteligncia Artificial <BR>
 * Sistemas de Informao - 01/2009 <BR>
 * Professor: Guilherme Dhein <BR>
 * <BR>
 * @author Douglas Pasqualin
 * @author Reinaldo Dornelles
 *
 */
public class Populacao 
{
	private Vector<Cromossomo> populacao;
	private int totalCidades;
	private Custos custos;
	private int totalMutacoes;
	private int totalSexuadas;
	private int fitnessTotal;
	private double chanceTotal;
	
	/**
	 * Mtodo construtor, instancia a classe e gera a populao inicial
	 * @param tamanhoPopulacao Indica quantos indivduos possui a populao
	 * @param totalCidades total de cidades que estaro na rota para o caixeiro viajar
	 * @param custos objeto do tipo Custos com os custos de deslocamento de todas as cidades
	 */	
	public Populacao(int tamanhoPopulacao, int totalCidades, Custos custos)
	{
		this.populacao = new Vector<Cromossomo>();
		this.totalCidades = totalCidades;
		this.custos = custos;
		this.totalSexuadas = (int) (tamanhoPopulacao * 0.7); 
		//70% para reproduo sexuada, importante que seja Par o valor
		if (totalSexuadas % 2 != 0)
			totalSexuadas++;
		//1 vai ser assexuada, o resto mutao
		this.totalMutacoes = tamanhoPopulacao - (totalSexuadas + 1);
		geraPopulacao(tamanhoPopulacao);
	}
	
	/**
	 * Mtodo utilizado a cada ciclo para gerar uma nova populao. <BR>
	 * A populao  gerada com reprodues: Assexuada, Mutao e Sexuada
	 */
	public void geraNovaPopulacao()
	{
		Reproducao reproducao = new Reproducao(totalCidades);
		Vector<Cromossomo> novaPopulacao = new Vector<Cromossomo>();
		//Reproducao assexuada, elitismo
		novaPopulacao.add(reproducao.reproducaoAssexuada(getMelhorAvaliado()));
		//Mutacaoes
		for (int i=0; i < totalMutacoes; i++)
			novaPopulacao.add(reproducao.mutacao(getCromossomoByRoleta()));

		Cromossomo[] novos = new Cromossomo[2];				
		for (int i=0; i < totalSexuadas / 2; i++)
		{
			novos = reproducao.reproducaoSexuada(getCromossomosByRoleta());
			novaPopulacao.add(novos[0]);
			novaPopulacao.add(novos[1]);
		}
			
		//Substitui a populao antiga
		populacao.removeAllElements();
		populacao = novaPopulacao;
		calculaFitnessTotal();
	}
	
	/**
	 * Mtodo para gerar a populao inicial e calcular o fitness total.
	 * Chamado no construtor da classe.
	 * @param tamanhoPopulacao Indica quantos indivduos possui a populao
	 */
	private void geraPopulacao(int tamanhoPopulacao)
	{
		for (int i = 0; i< tamanhoPopulacao; i++)
			populacao.add(new Cromossomo(totalCidades, custos));
		calculaFitnessTotal();
	}
	
	/**
	 * Mtodo para buscar o melhor avaliado da populao atravs do fitness mais BAIXO
	 * @return objeto do tipo Cromossomo com todas as informaes do melhor avaliado
	 */
	public Cromossomo getMelhorAvaliado()
	{
		//assumo que o primeiro  o melhor de todos
		Cromossomo melhor = populacao.get(0);
		for (int i = 1; i< populacao.size(); i++)
			if (populacao.get(i).getFitness() < melhor.getFitness())
				melhor = populacao.get(i);  
		return melhor;
	}
	
	/**
	 * Calcula o fitness total da populao para aps poder calcular
	 * o fitness relativo (probabilidade de reproduo)
	 */
	private void calculaFitnessTotal()
	{
		this.fitnessTotal = 0;
		for (int i = 0; i< populacao.size(); i++)
			this.fitnessTotal += populacao.get(i).getFitness();
		calculaChancesReproducao();
	}

	/**
	 * Calcula a probabilidade de cada individuo ser reproduzido
	 */
	private void calculaChancesReproducao()
	{
		double chance;
		this.chanceTotal = 0;
		for (int i = 0; i< populacao.size(); i++)
		{
			chance = (double)this.fitnessTotal / (double)populacao.get(i).getFitness();
			chance = Math.pow(chance, 3); //Eleva ao cubo
			chanceTotal += chance; //Acumula chanceTotal para utilizar na roleta
			populacao.get(i).setChanceReproducao(chance);
		}
	}
	
	/**
	 * Escolhe um indivduo aleatoriamente baseado na probabilidade de se reproduzir. 
	 * Utilizado para Mutao
	 * @return objeto do tipo Cromossomo com todas as informaes do escolhido
	 */
	private Cromossomo getCromossomoByRoleta()
	{
		Random random = new Random();
		int roleta = random.nextInt((int) chanceTotal);
		int acum = 0;
		for (int i = 0; i< populacao.size(); i++) 
		{
			if (acum + populacao.get(i).getChanceReproducao() > roleta)
				return populacao.get(i);
			else
				acum += populacao.get(i).getChanceReproducao();
		}
		//S para garantir que sempre vai retonar algum, 
		//ltimo caso retorna o ltimo da populao
		return populacao.get(populacao.size()-1);
	}

	/**
	 * Escolhe DOIS indivduos aleatoriamente baseado na probabilidade de se reproduzir. 
	 * Mtodo garante que no sero escolhidos dois individuos iguais. 
	 * Utilizado para Reproduo Sexuada.
	 * @return vetor de duas posies de objetos do tipo Cromossomo com todas as informaes do escolhido
	 */
	private Cromossomo[] getCromossomosByRoleta()
	{
		Cromossomo[] novo = new Cromossomo[2];
		novo[0] = getCromossomoByRoleta();
		do
		{
			novo[1] = getCromossomoByRoleta();
		}
		while (novo[0].getFitness() == novo[1].getFitness());

		return novo;
	}	
	
	/**
	 * Mostra na tela o fitness de todos os indivduos da populao e no final
	 * mostra o fitness do melhor avaliado da populao.
	 * Mtodos utilizado para auxiliar na depurao do cdigo
	 * @deprecated
	 * utilizado somente para depurao do cdigo.
	 */
	public void mostrarFitnessTodos()
	{
		for (int i = 0; i< populacao.size(); i++)
			System.out.println(populacao.get(i).getFitness() + "\n");
		System.out.println("Melhor de todos :" + getMelhorAvaliado().getFitness());
	}
	
	/**
	 * Ordena a populao pela chance de reproduo. Usado na roleta
	 * @deprecated
	 * No precisa estar ordenado para usar a roleta
	 */
	public void ordenaPopulacaoByChance()
	{
		Collections.sort (populacao, new Comparator<Cromossomo>() {  
			public int compare(Cromossomo c1, Cromossomo c2) {
				return c1.getChanceReproducao() > c2.getChanceReproducao() ? 1 : -1;
			}  
		});  
	}
	
}
